import java.awt.*;
import java.awt.event.*;


public final class Gui extends Frame implements Runnable {

  private long     pausePerCycle = 0; //200;
    private int      numObjects = 80;// 100;  //1000; //previously 700;
  private boolean  verbose = false;

    private int      worldSize = 35; //55; //100; //50;
  private int      pixelsPerLetter = 17;
  private int      boxSize;
  private int      boxX = 10;
  private int      boxY = 10;
  private int      display = 30;
  private int      topIndent = boxY+20+display;
  private Graphics page;
  private char[][] cacheGrid;
  private Font     font = new Font("Times", Font.BOLD, 12);

  private World    world;
  private boolean  autoStep = false;  // must synchronize to use this variable


  public static void main(String[] args) {  
    Gui gui = new Gui(args);
    gui.run();
  }


  public Gui(String[] args) {
    getCommandLineOptions(args);
      
    System.out.println("verbose = " + verbose);
    System.out.println("Simulation running...");

    trace("world_size: " + worldSize);
    trace("num_objects: " + numObjects);
    trace("pause_per_cycle: " + pausePerCycle);

    boxSize = worldSize * pixelsPerLetter;
    setSize(boxX*2 + boxSize, boxY+topIndent+boxSize);
    setLocation(37,20);//added to reveal top border
    setVisible(true);

    // Anonymous inner-class listener to terminate program
    this.addWindowListener(
      new WindowAdapter(){  // anonymous class definition
        public void windowClosing(WindowEvent e) {
          System.exit(0);  // terminate the program
        }
      }
    );//end addWindowListener

    page = getGraphics();

    //     enableEvents(AWTEvent.WINDOW_EVENT_MASK);
    enableEvents(AWTEvent.MOUSE_EVENT_MASK); 

    // Create and populate the world
      
    trace("calling constructor World(" + worldSize + ")");
    world = new World(worldSize);

    trace("calling world.initialize()");
    world.populate(numObjects);

    // this caches the current display to update as little as possible
    cacheGrid = new char[worldSize][worldSize];

    for(int row=0;row<worldSize;row++) 
      for(int col=0;col<worldSize;col++) 
        cacheGrid[row][col] = ' ';
  }


  public void run() {
    long t0, t1, timeElapsed;
    while (true)  {
      waitForAutoStep();
      t0 = System.currentTimeMillis();
      step();
      t1 = System.currentTimeMillis();
      timeElapsed = t1 - t0;
      if (timeElapsed < pausePerCycle) {
        try {
          Thread.sleep(pausePerCycle - timeElapsed);
        } catch (InterruptedException e) {}
      }
    }
  }


  public void processMouseEvent(MouseEvent e) {
    switch(e.getID()) {
    case MouseEvent.MOUSE_PRESSED:
      // If you pressed the mouse, do the next generation...
      trace("MouseEvent.MOUSE_PRESSED");
      doSingleStep();
      break;
    case MouseEvent.MOUSE_ENTERED:
      // If mouse enters window, stop running
      trace("MouseEvent.MOUSE_ENTERED");
      setAutoStep(false);
      break;
    case MouseEvent.MOUSE_EXITED:
      // If mouse leaves window, go full speed
      trace("MouseEvent.MOUSE_EXITED");
      setAutoStep(true);
      break;
    }
    super.processMouseEvent(e);
  }


  // synchronized methods:

  private synchronized void setAutoStep(boolean b) {
    autoStep = b;
    if (autoStep) {
      notify();
    }
  }

  private synchronized void waitForAutoStep() {
    while (! autoStep) {
      try {
        wait();
      } catch (InterruptedException e) {}
    }
  }

  private synchronized void doSingleStep() {
    if (! autoStep) {
      step();
    }
  }


  private void step() {
    world.nextTimeStep();
    updateWorldDisplay();
  }
 

  public void paint(Graphics page) 
    {
      page.setFont(font);
      page.drawRect(boxX-2,topIndent-2,boxSize+4, boxSize+4); 
      
      if (world != null)
	drawWorldDisplay();
    }


  private void updateWorldDisplay()
    {
      char [] letterArray = new char[1];
      char letter;
      int x;
      int y;

      page.setFont(font);
      for(int row=0;row<worldSize;row++) {

	for(int col=0;col<worldSize;col++) {
	  letter = world.getChar(row,col);

	  // If the character in this spot changed!  Redraw it
	  if (letter != cacheGrid[row][col])
	    {
	      cacheGrid[row][col] = letter;

	      x = boxX+(col*pixelsPerLetter);
	      y = row*pixelsPerLetter+(topIndent+pixelsPerLetter);

	      // Blank it out
	      page.clearRect(x-1, y-pixelsPerLetter+3, pixelsPerLetter, pixelsPerLetter);
	      
	      letterArray[0] = letter;
	      
	      // Redraw it	 
	      page.drawChars(letterArray,0,1,x,y);
	    }
	}
      }
      updateTimeDisplay();
    }

  private void drawWorldDisplay()
    {
      char [] letterArray = new char[1];
      int x;
      int y;

      page.setFont(font);
      for(int row=0;row<worldSize;row++) {

	for(int col=0;col<worldSize;col++) {
	 letterArray[0] = world.getChar(row,col);

	  cacheGrid[row][col] = letterArray[0];

	  x = boxX+(col*pixelsPerLetter);
	  y = row*pixelsPerLetter+(topIndent+pixelsPerLetter);
	  
	  // Redraw it	 
	 try{
	   page.drawChars(letterArray,0,1,x,y);
	 }
	 catch (ArrayIndexOutOfBoundsException e)
	   {
	   }

	}
      }
      updateTimeDisplay();
    }
  
  private int timeX = 10;
  private int timeY = 50;

  private void updateTimeDisplay()
    {
      page.clearRect(timeX-1, timeY-pixelsPerLetter, pixelsPerLetter*10, pixelsPerLetter);
      page.drawString("Time: " + world.getCurrentTimeStep(),timeX,timeY);
    }


  private void getCommandLineOptions(String args[])
    {
      for (int i = 0; i < args.length; i++) {
	
	try {
	  
	  if (args[i].equals("-verbose")) {
	    verbose = true;
	  } 
	  else if (args[i].equals("-numObjects")) {
	    i++;
	    numObjects = Integer.parseInt(args[i]);
	  }
	  else if (args[i].equals("-worldSize")) {
	    i++;
	    worldSize = Integer.parseInt(args[i]);
	  }
	  else if (args[i].equals("-pausePerCycle")) {
	    i++;
	    pausePerCycle = Integer.parseInt(args[i]);
	  }
	  else {
	    usage();
            System.exit(0);
	  };
	}
	catch(Throwable e) {
	  System.out.println("Problem with command line arguments... \n");
	  usage();
          System.exit(1);
	}
      }
    }

  private void usage() {
    System.out.println("Usage: java Gui [options]");
    System.out.println();
    System.out.println(" Options: \n");
    System.out.println("   -verbose:   Prints debugging statements in xterm.\n");
    System.out.println("   -worldSize <n>: The size of the world <nxn> (default: 50)\n");
    System.out.println("   -numObjects <n>: The number of objects initially created on the");
    System.out.println("                     world. (default: 500)\n");
    System.out.println("   -pausePerCycle <n>: Length of pause between " + "calls to nextTimeStep.");
    System.out.println("                         (default: 0)\n");
    System.out.println("   -help: Prints this message.\n");
  }

  private void trace(String s) {
    if (verbose) {
      System.out.println(s);
    }
  }

} // end of class Gui
