/**
 * This class is the simulation object; Animal is the root class of 
 * the dweller hierarchy provided by the user
 *  worldGrid is the representation of the World object
 *  there are worldSize*worldSize points in the World
 *  current is the encoded direction of the ocean current
 *  time is the current time step
 *  kint sets the interval of steps on which reproduction is tried
 *  timeStep is a static variable to keep track of number of
 *  time steps so far in the World simulation.
 *
 * @author Matthew Arnold, extended by Barbara Ryder
 */
public class World extends Object
{
    private WorldGrid worldGrid;//World rep type as a grid
    private int worldSize;//size of square rep
    private int current;//as in ocean current:
    //+1 means W to E; -1 means E to W
    //any direction with x coord as +1 will be in current 
    //dirn for W to E; etc.
    private int time;// The current time step; 
    private int kint;//sets inteval in which reproduction is tried
    private static int timeStep; //number of time steps so far

    /**
     *   observer for worldGrid
     *  @return worldGrid object
     */
    public WorldGrid getWorldGrid() { return worldGrid; }

    /**
     *   observer for worldSize
     *  @return size of the simulation grid which is size by size
     */
    public int getWorldSize() { return worldSize; }

    /**
     *   observer for current
     * @return value of current (+1 or -1)
     */
    public int getCurrent(){ return current;}

    /**
     *  This method simply iterates through all objects in the World and 
     *  calls processTimeStepForObject(obj,loc) on each object
     *  Gui.main() calls this nextTimeStep function for each time step
     *  in the simulation
     */
  public void nextTimeStep()
  {
      //increment timeStep
	timeStep++;
	
      // Setup the iterator on World for this time step
      //has to return Animal object/Location pairs
      WorldEnumeration worldEnum = new WorldEnumeration(this);
      while(worldEnum.hasMoreElements()) 
      {	
	// Get the next Location/Object pair
	LocationObjectPair pair = worldEnum.nextElement();
	// Extract the info from the pair
	Animal obj = pair.getObject();
	Location loc = pair.getLocation();
	
    // We need to check whether the Animal we get from the
    // iterator is STILL in its spot.  It is possible that the 
    // world now has a different object in that spot
    // If its there, process it.  If not, skip it.
    //i think this is true for simulations where objects can 
    //consume other objects; for the Shore similulation, this is 
    //not the case, but this code doesn't hurt
	
	if (obj == worldGrid.getElem(loc)) 
	  processTimeStepForObject(obj,loc);
       }

      time++;//increments time because have iterated all over the world
   }
    
    /**
     *   This function takes as arguments an Animal and its location 
     *   in the World.  The job of this function is to determine whether 
     *   the object dies, moves, or reproduces and to perform the correct
     *   resulting actions. This method must be changed for different
     *   simulations
     *  @param obj Animal object to be acted upon
     *  @param loc Location of Animal
     */
  private void processTimeStepForObject(Animal obj, Location loc)
  {	//
        //on each time step, an Animal 
	//1. ages; if it hits its max age then it may die;
        //2. moves some distance (if it can move) 
        //3. reproduces, if possible (depends on timeStep and
	// probability of offspring survival
        //
 
      // -------  Age and Death ---------
      // a dying object does not move or reproduce
      //
      //System.out.println("BGR object=" + String.valueOf(obj.getPrintChar()));
      
      obj.incrAge();//increments age of object
   //System.out.println("BGR age is "+obj.getAge()+" maxAge "+obj.getmaxAge());
      
      if (obj.timeToDie()) {//checks if object dies on this timeStep
	        worldGrid.setElem(loc,null);
		//System.out.println("BGR timeToDie is true");
	        return;
      }
      
      // ----  Movement ------
      
      //make_Move() just returns a Location to move to 
      //in this veresion of World, makeMove() 
      //does not throw Exception NoPossibleMoveExc, but it could
      //to show there are no choices for a move
      //fast animals will search further from their current location
      //for possible moves than slower animals;
      //
	  //System.out.println ("BGR animal tries to move");

    	    Location newLoc = obj.makeMove(loc,this);
	    if (newLoc != null) {
		// check move is legal
		if (whatIs(newLoc) == null) 
                { //no animal on desired location
		  //Move the Animal
		    //System.out.println ("BGR animal can move");
		  worldGrid.setElem(newLoc,obj);
    		  worldGrid.setElem(loc,null);
		}
		
	    }//only moves if non-null value returned by makeMove() 
	    //move is done
      
      // --------  Reproduction ----------
      // Objects given chance to multiply every kint time steps
      //
      if (timeStep%kint !=0) return;
      
      if (obj instanceof ReproducesbyEggs) {
          //
	  //code ignores situations where students' code sends
	  //more offspring than there is room for
	  //in neighborhood
	  //
	  
		    ReproducesbyEggs re = (ReproducesbyEggs) obj;
	      	    Set offspring = re.laysEggs();
		    if (offspring != null) placeOffspring(offspring,loc);
	     //System.out.println("BGR repro by eggs" + offspring.isEmpty());

	    }     

       if (obj instanceof ReproducesbyBudding){
	    ReproducesbyBudding rb = (ReproducesbyBudding) obj;
            Set offspring = rb.buds();
	    if (offspring != null)  placeOffspring(offspring,loc);
	    //System.out.println("BGR repro by buds" + offspring.isEmpty());
	    }
}
   
    /**
     *   this function places the offspring of an Animal in its neighborhood
     *  @param offspring set of Animal objects
     *  @param loc location of parent Animal object
     */
   private void placeOffspring(Set offspring, Location loc)
    {
      //loop places all offspring created in neighborhood of loc, if possible

      DirectionIter dIter = new DirectionIter();
      Direction d = dIter.next();//get first direction to explore

      SetEnumeration se = new SetEnumeration(offspring);
      while (se.hasMoreElements() && d != null)
	{   Animal obj = (Animal)se.nextElement();
	    //skip spots already occupied
	    while (d != null && whatIs(loc.plus(d)) != null)
	    d=dIter.next();
	    if (d != null) worldGrid.setElem(loc.plus(d),obj);
        }
      //do no check to see if can't place all the offspring     
    }

    /**
     *  this method returns the character representation of the Animal 
     *   which is at position <row,col> 
     *  @param row is x co-ordinate
     *  @param col is y co-ordinate
     */
  public char getChar(int row,int col)
    {
      Animal obj = worldGrid.getElem(row,col);
      if (obj == null)	return ' ';
      else return obj.getPrintChar();
    }

    /**
     * constructor - creates worldGrid representation and 
     * initializes fields current, timeStep, kint
     *
     * @parm world_size sets size of worldGrid object
     */
  World(int world_size) 
    {
      worldSize = world_size;
      worldGrid = new WorldGrid(worldSize,worldSize);
      if (Math.random() < .5) current = +1 ;
      else current = -1;
      timeStep= 0;//keeps track of time in the simulation
      kint = 3;//reproduction interval in timeSteps
    }

  //keeps a count of the total number of objects placed in the world
  static int totalPlaced=0;
  
    /**
     *  this function places Animals in the World 
     *  called by Gui.main()
     *  needs to be rewritten for every simulation
     * @param numObjects is total number of dwellers to be created
     * in the simulation
     */
  public void populate(int numObjects)
    {  //half the objects last time were nonmoving

    while (totalPlaced < numObjects) {

	totalPlaced++;
	int x = (int)(Math.random() * worldSize);
	int y = (int)(Math.random() * worldSize);
	if (worldGrid.getElem(x,y) != null)
	  continue;
	  //debug output
	double rand = Math.random()*10;
	//System.out.println( "rand= "+ rand);
   //
   //need to substitute each simulations own objects here
   //also need to test these numbers to see what works.
   //not clear what proportion of movers to nonmovers makes 
   //a good simulation
   //	
	if (rand < .5)
	    worldGrid.setElem(x,y,new Crab());
	else if (rand < 1.)
	  worldGrid.setElem(x,y,new Starfish());
	else if (rand < 2.0)
	  worldGrid.setElem(x,y,new Shrimp());
	else if (rand < 3.5)
	  worldGrid.setElem(x,y,new Seahorse());
	else if (rand < 4.5)
	  worldGrid.setElem(x,y,new Oyster());
	else if (rand < 5.5)
	  worldGrid.setElem(x,y,new Barnacle());
	else if (rand < 7.5)
	  worldGrid.setElem(x,y,new Anemone());
	else 
	 worldGrid.setElem(x,y,new Hydroid());
    }
 }
    /**
     *  observer for time
     * @return time of simulation
     */
  public int getCurrentTimeStep()
    {
      return time;
    }
    /**
     *  function checks location loc in simulation
     *  @param loc is location in question
     *  @return Animal at location loc or null
     */
    public Animal whatIs(Location loc)
    {
      return worldGrid.getElem(loc);
    }
}
















