/**
 * This class provides a representation for the simulation (World)
 *
 * @author Matthew Arnold, extended by Barbara Ryder
 */
public final class WorldGrid
{

  private Animal [][] grid;
  private int rows;
  private int cols;

    /**
     *  constructor for WorldGrid object which has an array, grid, for
     * the Animal objects
     * @param r is number of rows
     * @param c is number of cols
     *
     */
  WorldGrid(int r, int c)
    {
      rows = r;
      cols = c;
      grid = new Animal[rows][cols];
    }

    /**
     *  maintains the toroidal shape of the world by wrapping-around
     *  ints to be used as location co-ordinates
     * @param i is integer value
     * @param is integer to be used in mod operation of wrap
     *
     * @return int value of i%max
     */
  final private int wrap(int i,int max)
    {
      if (i >= 0) {
	if (i < max)
	  return i;
	return i % max;
      }
      else {
	// i < 0
	return max + (i % max)-1; 
      }
    }

    /**
     *
     * observer of elements of WorldGrid
     * @param loc is location whose occupant is requested
     *
     * @return reference to Animal at Location loc; can be null
     */
  public final Animal getElem(Location loc)
    {
      return grid[wrap(loc.getX(),rows)][wrap(loc.getY(),cols)];
    }

    /**
     *
     * observer of elements of WorldGrid
     * @param x is x co-ord of location whose occupant is requested
     * @param y is y co-ord of location whose occupant is requested
     *
     * @return reference to Animal at location x,y; can be null
     */
  public final Animal getElem(int x, int y)
    {
      return grid[wrap(x,rows)][wrap(y,cols)];
    }

    /**
     *
     * mutator of WorldGrid; stores reference to object obj at Location loc
     * @param loc is location to be wrapped
     * @param obj is Animal reference to be stored at wrapped location
     *
     */
  public final void setElem(Location loc,Animal obj)
    {
      grid[wrap(loc.getX(),rows)][wrap(loc.getY(),cols)] = obj;
    }

    /**
     *
     * mutator of WorldGrid; stores reference to object obj at location x,y
     * @param x is x co-ordinate of location
     * @param y is y co-ordiante of location
     * @param obj is reference to be stored at location
     *
     */
  public final void setElem(int x, int y, Animal obj)
    {
      grid[wrap(x,rows)][wrap(y,cols)] = obj;
    }

    /**
     *
     * observer of WorldGrid; 
     *
     * @return returns LocationObjectPair built from  Animal at loc 
     * and Location loc
     * @param loc is location in simulation
     */
  public final LocationObjectPair getGridElement(Location loc)
    {
      Animal obj = getElem(loc);

      if (obj == null)
	return null;
      else
	return new LocationObjectPair(loc,obj);
    }

}
